<?php
namespace Controller;
use Service\ColumnsQueryService;
use Service\ConnectionService;
use Service\DataSourceService;
use Service\DeleteCascadeService;
use Service\Manager\Mongo\Manager;
use Util\Utilities;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DTO\Response\Message;
use Zeedhi\Framework\DTO;

class StructureType extends Crud{

    const ERRORS_CONNECT_FAIL = "PHP Error occurred 'oci_connect()";
    const ERRORS_EXECUTE_QUERY_FAIL = "PHP Error occurred 'oci_execute()";

    const RESPONSE_STRING_CONNECTION_NOT_FOUND = "Connection not found: ";
    const RESPONSE_STRING_FAIL_AT_CREATE = "Wasn't possible create the connection: ";
    const RESPONSE_STRING_CONNECTION_FAIL = "Wasn't possible connect with database, the old columns were kept.";
    const RESPONSE_STRING_EXECUTE_QUERY_FAIL = "Wasn't possible sync your columns. Fail to execute query: ";

    protected $dataSourceName = 'structureType';

    /** @var Connection */
    protected $connection;
    
    protected $dataSourceService;
    
    protected $dataSourceOtherService;
    
    protected $socketClient;
    
    protected $environment;

    /** @var ConnectionService */
    protected $connectionService;

    public function __construct(Manager $organizationMongoManager, DeleteCascadeService $deleteCascadeService, ConnectionService $connectionService, \Service\DataSourceService $dataSourceService, \Service\DataSource $dataSourceOtherService, \Service\SocketClient $socketClient, \Util\Environment $environment) {
        $this->connectionService = $connectionService;
        $this->socketClient = $socketClient;
        $this->dataSourceService = $dataSourceService;
        $this->dataSourceOtherService = $dataSourceOtherService;
        $this->environment = $environment;
        parent::__construct($organizationMongoManager, $deleteCascadeService);
    }

    public function find(DTO\Request\Filter $request, DTO\Response $response) {
        parent::find($request, $response);
        $dataSet = $response->getDataSets()[0];
        $structureTypes = $dataSet->getRows();
        $connections = $this->getConnections($structureTypes)->getRows();
        $indexedConnections = $this->indexByMongoId($connections);
        $hierarchyTypes = $this->relateStructureTypeWithConnection($structureTypes, $indexedConnections);
        $newDataSet = new DataSet($this->dataSourceName, $hierarchyTypes);
        $response->setDataSets(array($newDataSet));
    }

    protected function getConnections ($structureTypes) {
        $usedIds = $this->getConnectionsIdsUsed($structureTypes);
        $numberOfExpectedLines = count($usedIds);
        $filterCriteria = new FilterCriteria('connection', array(), null, $numberOfExpectedLines);
        $filterCriteria->addCondition('_id', FilterCriteria::IN, $usedIds);
        return $this->mongoManager->findBy($filterCriteria);
    }

    protected function getConnectionsIdsUsed ($structureTypes) {
        //For not repeat Ids used more than one time
        $connectionIdsIndexedById = $this->getDistinctConnectionsIds($structureTypes);
        //Build distinct ids in a index array
        $distictConnectionIds = array();
        foreach($connectionIdsIndexedById as $connectionId) {
            $distictConnectionIds[] = $connectionId;
        }
        return $distictConnectionIds;
    }

    protected function getDistinctConnectionsIds($structureTypes) {
        $connectionsIdsIndexedById = array();
        foreach($structureTypes as $structureType) {
            if (isset($structureType['connection'])) {
                $connectionId = $structureType['connection'];
                $connectionsIdsIndexedById[$connectionId] = $connectionId;
            }
        }
        return $connectionsIdsIndexedById;
    }

    protected function indexByMongoId($rows) {
        $newRows = array();
        foreach($rows as $row) {
            $newRows[$row['_id']] = $row;
        }
        return $newRows;
    }

    protected function relateStructureTypeWithConnection($structureTypes, $indexedByIdConnections) {
        foreach($structureTypes as &$structureType) {
            if (isset($structureType['connection']) && isset($indexedByIdConnections[$structureType['connection']])) {
                $structureType['connectionName'] = $indexedByIdConnections[$structureType['connection']]['name'];
            }
        }
        return $structureTypes;
    }

    public function getFieldKeysFromQuery(DTO\Request $request, DTO\Response $response) {
        $request = $this->buildDataSetIfRow($request);
        $row = $request->getDataSet()->getRows()[0];
        try {
			//TODO: use custom connection if there is one (first it need to be implemented in the websocket server)
            $columns = $this->getColumnsFromWebSocket($row);
            $response->addDataSet(new DataSet('columns', $columns));
        } catch (Exception $e) {
            if($e->getCode() === Exception::CODE_FAIL_CONNECTION || $e->getCode() === Exception::CODE_QUERY_EXECUTION_FAIL) {
                $response->addMessage(new Message($e->getMessage(), Message::TYPE_ERROR));
            } else {
                throw $e;
            }
        }
    }
    
    protected function getColumnsFromWebSocket($row) {
        try {
            $cleanedQuery = $this->dataSourceOtherService->replaceQueryParameters($row['query']);
            $organizationID = Utilities::getOrganizationIdForMongo();
            return $this->dataSourceService->getMergedColumnsFromWebSocket($this->socketClient, $cleanedQuery, $row, $organizationID);
        } catch (\Zeedhi\Framework\ErrorHandler\ErrorException $e) {
            if (strrpos($e->getMessage(), self::ERRORS_CONNECT_FAIL)) {
                throw Exception::connectFail();
            }
            else if (strrpos($e->getMessage(), self::ERRORS_EXECUTE_QUERY_FAIL)) {
                throw Exception::executeQueryError($row['query'], $e);
            }
            else {
                throw $e;
            }
        }
    }
}